package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import javax.vecmath.Vector3d;

/**
 * Finds and stores the top X points with the most significant curvarure.
 * 
 * @author Radek Oslejsek
 * @author Natalia Bebjakova
 */
public class SignificantPoints {

    private final List<VertCurvature> significantPoints;
    
    private List<List<Vector3d>> normCosVecCache;
    
    private List<List<Vector3d>> avgPosCache;
    
    private List<List<Double>> curRatioCache;
    
    /**
     * Constructor.
     * @param curvatures List of points' curvatures
     * @param max Maximal number of points
     */
    public SignificantPoints(List<Double> curvatures, int max) {
        List<VertCurvature> sp = new ArrayList<>(curvatures.size());
        for (int i = 0; i < curvatures.size(); i++) { 
            sp.add(new VertCurvature(i, curvatures.get(i)));
        }
        
        Collections.sort(sp);
        if (max < sp.size()) {
            significantPoints = new ArrayList<>(sp.subList(0, max));
        } else {
            significantPoints = sp;
        }
        
    }
    
    /**
     * Precomputes and caches some values related to the computation of
     * asymmetric plane from curvature.
     * 
     * @param facet Mesh facet
     */
    public void cacheData(MeshFacet facet) {
        if (normCosVecCache != null) {
            return;
        }
        
        normCosVecCache = new ArrayList<>(significantPoints.size());
        avgPosCache = new ArrayList<>(significantPoints.size());
        curRatioCache = new ArrayList<>(significantPoints.size());
        
        for (int i = 0; i < significantPoints.size(); i++) {
            List<Vector3d> cosArray = new ArrayList<>(significantPoints.size());
            normCosVecCache.add(cosArray);
            
            List<Vector3d> posArray = new ArrayList<>(significantPoints.size());
            avgPosCache.add(posArray);
            
            List<Double> curArray = new ArrayList<>(significantPoints.size());
            curRatioCache.add(curArray);
            
            for (int j = 0; j < significantPoints.size(); j++) {
                MeshPoint meshPointI = facet.getVertex(significantPoints.get(i).vertIndex);
                MeshPoint meshPointJ = facet.getVertex(significantPoints.get(j).vertIndex);
                
                Vector3d ni = new Vector3d(meshPointI.getNormal());
                Vector3d nj = new Vector3d(meshPointJ.getNormal());
                ni.normalize();
                nj.normalize();
                ni.sub(nj);
                ni.normalize();
                cosArray.add(ni);
                
                Vector3d avrg = new Vector3d(meshPointI.getPosition());
                Vector3d aux = new Vector3d(meshPointJ.getPosition());
                avrg.add(aux);
                avrg.scale(0.5);
                posArray.add(avrg);
                curArray.add(significantPoints.get(i).curvature / significantPoints.get(j).curvature);
            }
        }
    }
    
    /**
     * Returns cached normal vector for cos.
     * @param i index of the i-th significant curvature point
     * @param j index of the j-th significant curvature point
     * @return cached value or null
     */
    public Vector3d getCachedNormCosVec(int i, int j) {
        if (normCosVecCache == null || i >= normCosVecCache.size() || i >= normCosVecCache.size() || i < 0 || j < 0) {
            return null;
        }
        return normCosVecCache.get(i).get(j);
    }
    
    /**
     * Returns cached vector for the computation average normal.
     * @param i index of the i-th significant curvature point
     * @param j index of the j-th significant curvature point
     * @return cached value or null
     */
    public Vector3d getCachedAvgPos(int i, int j) {
        if (avgPosCache == null || i >= avgPosCache.size() || i >= avgPosCache.size() || i < 0 || j < 0) {
            return null;
        }
        return avgPosCache.get(i).get(j);
    }
    
    /**
     * Returns cached curvature ratio.
     * @param i index of the i-th significant curvature point
     * @param j index of the j-th significant curvature point
     * @return cached value or {@code Double.NaN}
     */
    public double getCachedCurRatio(int i, int j) {
        if (curRatioCache == null || i >= curRatioCache.size() || i >= curRatioCache.size() || i < 0 || j < 0) {
            return Double.NaN;
        }
        return curRatioCache.get(i).get(j);
    }
    
    /**
     * Returns index of the i-th most significant mesh vertex.
     * 
     * @param i The "i" (index of the ordering)
     * @return index of the i-th most significant mesh vertex.
     * @throws IllegalArgumentException if the {@code i} paramter is out of rage
     */
    public int getVertexIndex(int i) {
        if (i >= 0 && i < significantPoints.size()) {
            return significantPoints.get(i).vertIndex;
        } else {
            throw new IllegalArgumentException("i");
        }
    }
    
    /**
     * Returns curvature of the i-th most significant mesh vertex.
     * 
     * @param i The "i" (index of the ordering)
     * @return curvature of the i-th most significant mesh vertex.
     * @throws IllegalArgumentException if the {@code i} paramter is out of rage
     */
    public double getCurvature(int i) {
        if (i >= 0 && i < significantPoints.size()) {
            return significantPoints.get(i).curvature;
        } else {
            throw new IllegalArgumentException("i");
        }
    }
    
    /**
     * Returns numnber of the most significant points.
     * @return numnber of the most significant points.
     */
    public int size() {
        return significantPoints.size();
    }
    
    /**
     * Helper class for sorting points with respect to their curvature.
     * @author Radek Oslejsek
     */
    private class VertCurvature implements Comparable<VertCurvature> {
        
        private final int vertIndex;
        private final double curvature;
        
        VertCurvature(int vertIndex, double curvature) {
            this.vertIndex = vertIndex;
            this.curvature = curvature;
        }

        @Override
        public int compareTo(VertCurvature arg) {
            return -Double.compare(this.curvature, arg.curvature);
        }
        
        public String toString() {
            return ""+curvature;
        }
    }
    
}
