package cz.fidentis.analyst.symmetry;

import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshFacetImpl;
import cz.fidentis.analyst.mesh.core.MeshPointImpl;
import cz.fidentis.analyst.visitors.mesh.BoundingBox;
import cz.fidentis.analyst.visitors.mesh.BoundingBoxVisitor;
import cz.fidentis.analyst.visitors.mesh.CurvatureVisitor;
import cz.fidentis.analyst.visitors.mesh.GaussianCurvatureVisitor;
import cz.fidentis.analyst.visitors.mesh.MaxCurvatureVisitor;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.vecmath.Vector3d;

/**
 * Main class for computing approximate plane of symmetry of the 3D model.
 * Default values of the configuration are given due to the best results on tested objects.
 * On many different 3D models, it exists other values of config that will have better impact on result. 
 * <p>
 * Note: this algorithm can be implemented as a mesh visitor because it requires 
 * complete mesh to compute the symmetry plane. Sequential application on 
 * multiple meshes does not work.
 * </p>
 * 
 * @author Natalia Bebjakova
 * @author Radek Oslejsek
 */
public class SymmetryEstimator {
    
    private final MeshFacet facet;
    private final Config config;
    private final boolean maxCurvatureAlg;
    
    // results:
    private List<Double> curvatures;
    private BoundingBox boundingBox;
    private Plane symmetryPlane;
    
    /**
     * Constructor.
     * 
     * @param facet Mesh facet for which the symmetry plane is calculated
     * @param config Algorighm options
     * @param maxCurvatureAlg If {@code true}, then more precise but slower {@link MaxCurvatureVisitor}
     * algorithm is used. Otherwise, the faster {@link GaussianCurvatureVisitor} is used. 
     * See {@link CurvatureVisitor} for more details.
     * @throws IllegalArgumentException if some input paramter is missing
     */
    public SymmetryEstimator(MeshFacet facet, Config config, boolean maxCurvatureAlg) {
        if (facet == null) {
            throw new IllegalArgumentException("facet");
        }
        if (config == null) {
            throw new IllegalArgumentException("config");
        }
        this.facet = facet;
        this.config = config;
        this.maxCurvatureAlg = maxCurvatureAlg;
    }
    
    /**
     * Helper constructor.
     * @param centroid Centroid
     * @param a A
     * @param b B
     * @param scale Scale
     */
    /*
    protected SymmetryEstimator(Vector3d centroid, Vector3d a, Vector3d b, double scale) {
        this(createMeshFacet(centroid, a, b, scale), Config.getDefault());
    }
    */
    
    /**
     * Calculates the symmetry plane.
     */
    public void calculateSymmetryPlane() {
               
        if (!facet.hasVertexNormals()) {
            facet.calculateVertexNormals();
        }
        
        curvatures = calculateCurvatures(facet, maxCurvatureAlg);
        final SignificantPoints sigPoints = new SignificantPoints(curvatures, config.getSignificantPointCount());
        sigPoints.cacheData(facet);
        
        // Initiate structure for concurrent computation:
        ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        final List<Future<ApproxSymmetryPlane>> results = new ArrayList<>(sigPoints.size()*sigPoints.size());
        
        // Compute candidate planes concurrently:
        final double maxDistance = calculateMaxRelativeDistance(facet, config);
        for (int i = 0; i < sigPoints.size(); i++) {
            for (int j = 0; j < sigPoints.size(); j++) {
                int finalI = i;
                int finalJ = j;
                
                Callable<ApproxSymmetryPlane> worker = new  Callable<ApproxSymmetryPlane>() {
                    @Override
                    public ApproxSymmetryPlane call() throws Exception {
                        try {
                            return new ApproxSymmetryPlane(facet, sigPoints, config, finalI, finalJ, maxDistance);
                        } catch(UnsupportedOperationException ex) {
                            return null;
                        }
                    }
                };
                
                results.add(executor.submit(worker));
            }
        }
        
        // Wait until all symmetry planes are computed:
        executor.shutdown();
        while (!executor.isTerminated()){
        }
        
        // Get the best-fitting planes:
        List<Plane> planes = new ArrayList<>();
        try {
            int maxVotes = 0;
            for (Future<ApproxSymmetryPlane> res: results) {
                ApproxSymmetryPlane newPlane = res.get();
                if (newPlane == null) {
                    continue;
                }
                if (newPlane.getVotes() > maxVotes) {
                    maxVotes = newPlane.getVotes();
                    planes.clear();
                    planes.add(newPlane);
                } else if (newPlane.getVotes() == maxVotes) {
                    planes.add(newPlane);
                }
            }
        } catch (final InterruptedException | ExecutionException ex) {
            Logger.getLogger(SymmetryEstimator.class.getName()).log(Level.SEVERE, null, ex);
        }
        
        if (config.isAveraging() && !planes.isEmpty()) {
            symmetryPlane = new Plane(planes);
        } else {
            symmetryPlane = planes.isEmpty() ? null : planes.get(0);
        }
    }
    
    /**
     * Returns the symmetry plane computed by the {@link SymmetryEstimator#calculateSymmetryPlane()} method.
     * @return the symmetry plane or {@code null}
     */
    public Plane getSymmetryPlane() {
        return symmetryPlane;
    }
    
    /**
     * COmputes and returns a triangular mesh for the symmetry plane (a rectangle). 
     * Its size is restricted by the bounding box of the original mesh facet.
     * 
     * @return mesh facet of the symmetry plane or {@code null}
     */
    public MeshFacet getSymmetryPlaneMesh() {
        if (symmetryPlane == null) {
            return null;
        }
        
        Vector3d normal = symmetryPlane.getNormal();
        Vector3d midPoint = boundingBox.getMidPoint().getPosition();

        double alpha = -((normal.x * midPoint.x) + 
                (normal.y * midPoint.y) + (normal.z * midPoint.z) +
                symmetryPlane.getDistance()) / (normal.dot(normal));
        
        Vector3d midPointOnPlane = new Vector3d(midPoint);
        Vector3d nn = new Vector3d(normal);
        nn.scale(alpha);
        midPointOnPlane.add(nn);

        double val = normal.x * midPointOnPlane.x + normal.y *
                midPointOnPlane.y + normal.z *
                midPointOnPlane.z + symmetryPlane.getDistance();

        Vector3d a = new Vector3d();
        if (Math.abs(normal.dot(new Vector3d(0.0, 1.0, 0.0))) > Math.abs(normal.dot(new Vector3d (1.0, 0.0, 0.0)))) {
            a.cross(normal, new Vector3d(1.0, 0.0, 0.0));
        } else {
            a.cross(normal, new Vector3d(0.0, 1.0, 0.0));
        }
        a.normalize();

        Vector3d b = new Vector3d();
        b.cross(normal,a);
        b.normalize();
        
        double scale = (boundingBox.getMaxPoint().subtractPosition(boundingBox.getMinPoint())).getPosition().x;
        
        return createMeshFacet(midPointOnPlane, a, b, symmetryPlane.getNormal(), scale);
    }
    
    /**
     * Returns the bounding box computed during the {@link SymmetryEstimator#calculateSymmetryPlane()}.
     * @return the bounding box or {@code null}
     */
    public BoundingBox getBoundingBox() {
        return boundingBox;
    }
    
    /**
     * Returns curvatures computed during the {@link SymmetryEstimator#calculateSymmetryPlane()}.
     * @return the curvatures or {@code null}
     */
    public List<Double> getCurvature() {
        return Collections.unmodifiableList(curvatures);
    }

    protected double calculateMaxRelativeDistance(MeshFacet facet, Config config) {
        BoundingBoxVisitor visitor = new BoundingBoxVisitor(false);
        facet.accept(visitor);
        boundingBox = visitor.getBoundingBox();
        return boundingBox.getMaxDiag() * config.getMaxRelDistance();
    }

    protected static List<Double> calculateCurvatures(MeshFacet facet, boolean maxCurvature) {
        CurvatureVisitor vis = (maxCurvature) ? new MaxCurvatureVisitor(false) : new GaussianCurvatureVisitor(false);
        facet.accept(vis);
        List<Double> curvatures = new ArrayList<>(vis.getCurvatures().get(facet));
        for (int i = 0; i < curvatures.size(); i++) {
            if (Double.isNaN(curvatures.get(i))) {
                curvatures.set(i, 0.0);
            }
        }
        return curvatures;
    }
    
    protected static MeshFacet createMeshFacet(Vector3d centroid, Vector3d a, Vector3d b, Vector3d normal, double scale) {
        Vector3d[] points = new Vector3d[4];
        
        Vector3d aScaled = new Vector3d(a);
        Vector3d bScaled = new Vector3d(b);
        aScaled.scale(scale);
        bScaled.scale(scale);
        
        points[0] = new Vector3d(centroid);
        points[0].sub(aScaled);
        points[0].sub(bScaled);
        
        points[1] = new Vector3d(centroid);
        points[1].sub(aScaled);
        points[1].add(bScaled);
        
        points[2] = new Vector3d(centroid);
        points[2].add(aScaled);
        points[2].add(bScaled);
        
        points[3] = new Vector3d(centroid);
        points[3].add(aScaled);
        points[3].sub(bScaled);
        
        MeshFacet facet = new MeshFacetImpl();
        for (Vector3d point : points) {
            facet.addVertex(new MeshPointImpl(point, normal, null));
        }
        //facet.calculateVertexNormals();
        
        return facet;
    }
    
    /**
     * 
     * @param s mesh that will be merged
     * @return mesh with merged vertices from both meshes 
     */
    /*
    protected static SymmetryEstimator mergeMeshWith(SymmetryEstimator s) {
        CornerTableRow row1 = new CornerTableRow(facet.getNumberOfVertices(), -1);
        CornerTableRow row2 = new CornerTableRow(facet.getNumberOfVertices() + 1, facet.getNumberOfVertices() + 3);
        CornerTableRow row3 = new CornerTableRow(facet.getNumberOfVertices() + 2, -1);
        CornerTableRow row4 = new CornerTableRow(facet.getNumberOfVertices() + 2, -1);
        CornerTableRow row5 = new CornerTableRow(facet.getNumberOfVertices() + 3, facet.getNumberOfVertices() + 1);
        CornerTableRow row6 = new CornerTableRow(facet.getNumberOfVertices(), -1);
        
        facet.getCornerTable().addRow(row1);
        facet.getCornerTable().addRow(row2);
        facet.getCornerTable().addRow(row3);
        facet.getCornerTable().addRow(row4);
        facet.getCornerTable().addRow(row5);
        facet.getCornerTable().addRow(row6);

        for(int n = 0; n < 4; n++) {
            facet.addVertex(s.getFacet().getVertices().get(n));
        }        
        return this;
    }
    */
}
