package cz.fidentis.analyst.visitors.kdtree;

import cz.fidentis.analyst.kdtree.KdNode;
import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.Queue;
import java.util.Set;
import javax.vecmath.Vector3d;

/**
 * This visitor finds <b>all nodes</b> that are closest to the given 3D location in a KD tree.
 * The minimal distance is computed between the 3D point and nodes of the KD tree.
 * <p>
 * The visitor returns the closest nodes and their absolute distance to the 3D point.
 * However, a single visitor can be applied to multiple KD trees.
 * <ul>
 * <li>If one tree has a closer distance to the given 3D point than all other trees,
       then only the closest nodes of that particular tree are returned.</li>
 * <li>If more than one tree share the same minimum distance to the given 3D point,
       the closest nodes of all those trees are returned.</li>
 * </ul>
 * </p>
 * <p>
 * This visitor is thread-safe. A single instance of the visitor can be used 
 * to inspect multiple KD trees simultaneously.
 * </p>
 * 
 * @author Daniel Schramm
 */
public class ClosestNodeMultipleVisitor extends ClosestNodeAbstractVisitor {
    
    private final Set<KdNode> closest = new HashSet<>();
    
    /**
     * @param point The 3D location for which the closest nodes are to be computed.
     * @param asynchronous If {@code true}, then asynchronous visitor is created.
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public ClosestNodeMultipleVisitor(Vector3d point, boolean asynchronous) {
        super(point, asynchronous);
    }

    /**
     * Returns set of the closest nodes of the visited KD tree.
     * 
     * @return closest nodes of the visited KD tree
     */
    public Set<KdNode> getClosestNodes() {
        return Collections.unmodifiableSet(closest);
    }

    @Override
    protected void visitKdTree(KdTree kdTree) {
        final KdNode root = kdTree.getRoot();
        if (root == null) {
            return;
        }
        
        KdNode foundNode = closestNode(kdTree.getRoot());
        final double closestDistance = MeshPoint.distance(getPoint3d(), foundNode.getLocation());
        synchronized (this) {
            if (updateMinimalDistance(closestDistance)) {
                closest.clear();
            }
        }
        
        // There is no closer point than the 'closestFound' in our tree.
        // Search for other closest points with the same distance.
        while (foundNode != null) {
            final double distanceFound = MeshPoint.distance(getPoint3d(), foundNode.getLocation());
            synchronized (this) {
                if (distanceFound == getDistance()) {
                    closest.add(foundNode);
                } else { // There is a closer point found in some other tree visited by this visitor
                    return;
                }
            }
            foundNode = closestNodeOmitting(root);
        }
    }
    
    /*
     * Find an undiscovered node with its distance equal to 'distance'.
     *
     * - Discovered nodes are those stored in the 'closest' set.
     * - We are not looking for distance closer than 'distance' because
     *   from the method {@link cz.fidentis.analyst.visitors.kdtree.ClosestNodeMultipleVisitor#visitKdTree}
     *   we already know distance to the closest node of the visited tree.
     */
    private KdNode closestNodeOmitting(KdNode root) {
        if (root == null) {
            return null;
        }
        
        /*
         * First, find the closest node
         */
        KdNode nearNode = null;
        KdNode searchedNode = root;

        while (searchedNode != null) {
            final Vector3d nodePos = searchedNode.getLocation();
            final double dist = MeshPoint.distance(getPoint3d(), nodePos);
            synchronized (this) {
                if (dist == getDistance() && !closest.contains(searchedNode)) {
                    nearNode = searchedNode;
                }
            }

            if (firstIsLessThanSecond(getPoint3d(), nodePos, searchedNode.getDepth())) {
                searchedNode = searchedNode.getLesser();
            } else {
                searchedNode = searchedNode.getGreater();
            }
        }

        /*
         * Second, search for a vertex that could be potentially closer than
         * the nearest vertex already found
         */
        final Queue<KdNode> queue = new LinkedList<>();
        queue.add(root);

        while (!queue.isEmpty()) {
            searchedNode = queue.poll();
            Vector3d nodePos = searchedNode.getLocation();

            final double dist = MeshPoint.distance(getPoint3d(), nodePos);
            synchronized (this) {
                if (dist == getDistance() && !closest.contains(searchedNode)) {
                    nearNode = searchedNode;
                }
            }

            final double distOnAxis = minDistanceIntersection(nodePos, getPoint3d(), searchedNode.getDepth());

            if (distOnAxis > getDistance()) {
                if (firstIsLessThanSecond(getPoint3d(), nodePos, searchedNode.getDepth())) {
                    if (searchedNode.getLesser() != null) {
                        queue.add(searchedNode.getLesser());
                    }
                } else {
                    if (searchedNode.getGreater() != null) {
                        queue.add(searchedNode.getGreater());
                    }
                }
            } else {
                if (searchedNode.getLesser() != null) {
                    queue.add(searchedNode.getLesser());
                }
                if (searchedNode.getGreater() != null) {
                    queue.add(searchedNode.getGreater());
                }
            }
        }

        return nearNode;
    }
    
}
