package cz.fidentis.analyst.visitors.mesh;

import cz.fidentis.analyst.kdtree.KdTree;
import cz.fidentis.analyst.kdtree.KdTreeVisitor;
import cz.fidentis.analyst.mesh.MeshVisitor;
import cz.fidentis.analyst.mesh.core.MeshFacet;
import cz.fidentis.analyst.mesh.core.MeshModel;
import cz.fidentis.analyst.mesh.core.MeshPoint;
import cz.fidentis.analyst.visitors.Distance;
import cz.fidentis.analyst.visitors.DistanceWithNearestPoints;
import cz.fidentis.analyst.visitors.kdtree.KdTreeApproxDistanceToTriangles;
import cz.fidentis.analyst.visitors.kdtree.KdTreeDistance;
import cz.fidentis.analyst.visitors.kdtree.KdTreeDistanceToVertices;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.vecmath.Vector3d;

/**
 * Visitor for Hausdorff distance. 
 * This visitor is instantiated with a single k-d tree (either given as the input
 * parameter, or automatically created from the triangular mesh). 
 * When applied to other mesh facets, it computes Huasdorff distance from them to 
 * the instantiated k-d tree.
 * <p>
 * This visitor is thread-safe.
 * </p>
 * <p>
 * The distance is computer aither as absolute or relative. Abosolute
 * represents Euclidean distance (all numbers are positive). On the contrary, 
 * relative distance considers orientation of the visited mesh (determined by its normal vectors)
 * and produces positive or negative distances depending on whether the primary
 * mesh is "in front of" or "behind" the given vertex.
 * </p>
 * <p>
 * Comparison of methods' efficiency performed with scans of two human faces on 
 * Intel(R) Core(TM) i7-8565U CPU @ 1.80GHz, 8 CPU cores.
 * </p>
 * <p>
 * Hausdorff distance sequentially:
 * <table>
 * <tr><td>2076 msec</td><td>POINT_TO_POINT_ABSOLUTE</td></tr>
 * <tr><td>3991 msec</td><td>POINT_TO_POINT</td></tr>
 * <tr><td>5528 msec</td><td>POINT_TO_TRIANGLE_APPROXIMATE</td></tr>
 * </table>
 * </p>
 * <p>
 * Hausdorff distance consurrently:
 * <table>
 * <tr><td>1583 msec</td><td>POINT_TO_POINT_ABSOLUTE</td></tr>
 * <tr><td>2399 msec</td><td>POINT_TO_POINT</td></tr>
 * <tr><td>2745 msec</td><td>POINT_TO_TRIANGLE_APPROXIMATE</td></tr>
 * </table>
 * </p>
 * 
 * @author Daniel Schramm
 * @author Radek Oslejsek
 */
public class HausdorffDistance extends MeshVisitor  {
    
    /**
     * Distance computation strategies.
     * @author Radek Oslejsek
     */
    public enum Strategy {
        /**
         * The fastest algorithm. Computes abolute distance between mesh vertices.
         * Relative distance is not supported.
         */
        POINT_TO_POINT_ABSOLUTE,
        
        /**
         * A bit slower point-to-point algorithm. 
         * But supports both relative and absolute distances.
         * However, for relative distance can produce noise. If there are multiple 
         * closest points to the given vertex, then 
         * {@code Double.POSITIVE_INFINITY} value is set as the distance.
         * If this situation happens with human faces, then we consider this part
         * of the face as noisy anyway.
         */
        POINT_TO_POINT,
        
        /**
         * The fast point-to-triangle distance strategy, which more precise
         * than point-to-point strategies. 
         * Supports both relative and absolute distances.
         * However, the algorithm is not geometrically correct. 
         * Can produce noise (does not found the really closests triangle) on wrinkly sufaces. 
         * Nevertheless, we aim to omit such noisy parts of human face from analysis
         * rather than dealing with this situation.
         */
        POINT_TO_TRIANGLE_APPROXIMATE
    }
    
    /**
     * Result. 
     * Key = visited mesh facets. 
     * Vvalue = Hausdorff distance of each vertex of the visited facet to the source facet.
     */
    private final Map<MeshFacet, List<Double>> distances = new HashMap<>();
    
    private final Strategy strategy;
    
    private final boolean relativeDist;
    
    private final boolean parallel;
    
    /**
     * The source triangular mesh (set of mesh facets) stred in k-d tree.
     */
    private final KdTree kdTree;
    
    /**
     * Constructor.
     * 
     * @param mainFacets Facets to which distance from other facets is to be computed. Must not be {@code null}
     * @param strategy Strategy of the computation of distance
     * @param relativeDistance If true, then the visitor calculates the relative distances with respect 
     * to the normal vectors of source facets (normal vectors have to present), 
     * i.e., we can get negative distances.
     * @param asynchronous If {@code true}, then the inspection of mesh facets is asynchronous (and possibly concurrent).
     * @param parallel If {@code true}, then the algorithm runs concurrently utilizing all CPU cores
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public HausdorffDistance(Set<MeshFacet> mainFacets, Strategy strategy, boolean relativeDistance, boolean asynchronous, boolean parallel) {
        super(asynchronous);
        this.strategy = strategy;
        this.relativeDist = relativeDistance;
        this.parallel = parallel;
        if (mainFacets == null) {
            throw new IllegalArgumentException("mainFacets");
        }
        if (strategy == Strategy.POINT_TO_POINT_ABSOLUTE && relativeDistance) {
            throw new IllegalArgumentException("Point-to-point strategy cannot be used with relative distance");
        }
        this.kdTree = new KdTree(new ArrayList<>(mainFacets));
    }

    /**
     * Constructor.
     * 
     * @param mainFacet Primary facet to which distance from others is to be computed. Must not be {@code null}
     * @param strategy Strategy of the computation of distance
     * @param relativeDistance If true, then the visitor calculates the relative distances with respect 
     * to the normal vectors of source facets (normal vectors have to present), 
     * i.e., we can get negative distances.
     * @param asynchronous If {@code true}, then the inspection of mesh facets is asynchronous (and possibly concurrent).
     * @param parallel If {@code true}, then the algorithm runs concurrently utilizing all CPU cores
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public HausdorffDistance(MeshFacet mainFacet, Strategy strategy, boolean relativeDistance, boolean asynchronous, boolean parallel) {
        this(new HashSet<>(Collections.singleton(mainFacet)), strategy, relativeDistance, asynchronous, parallel);
        if (mainFacet == null) {
            throw new IllegalArgumentException("mainFacet");
        }
    }

    /**
     * Constructor.
     * 
     * @param mainModel The mesh model with primary facets to which distance from 
     * others is to be computed. Must not be {@code null} or empty.
     * @param strategy Strategy of the computation of distance
     * @param relativeDistance If true, then the visitor calculates the relative distances with respect 
     * to the normal vectors of source facets (normal vectors have to present), 
     * i.e., we can get negative distances.
     * @param asynchronous If {@code true}, then the inspection of mesh facets is asynchronous (and possibly concurrent).
     * @param parallel If {@code true}, then the algorithm runs concurrently utilizing all CPU cores
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public HausdorffDistance(MeshModel mainModel, Strategy strategy, boolean relativeDistance, boolean asynchronous, boolean parallel) {
        this(new HashSet<>(mainModel.getFacets()), strategy, relativeDistance, asynchronous, parallel);
        if (mainModel.getFacets().isEmpty()) {
            throw new IllegalArgumentException("mainModel");
        }
    }
    
    /**
     * Constructor.
     * 
     * @param mainKdTree The KD tree to which distance from the visited facets is to be computed. Must not be {@code null}
     * @param strategy Strategy of the computation of distance
     * @param relativeDistance If true, then the visitor calculates the relative distances with respect 
     * to the normal vectors of source facets (normal vectors have to present), 
     * i.e., we can get negative distances.
     * @param asynchronous If {@code true}, then the inspection of mesh facets is asynchronous (and possibly concurrent).
     * @param parallel If {@code true}, then the algorithm runs concurrently utilizing all CPU cores
     * @throws IllegalArgumentException if some parametr is wrong
     */
    public HausdorffDistance(KdTree mainKdTree, Strategy strategy, boolean relativeDistance, boolean asynchronous, boolean parallel) {
        super(asynchronous);
        this.strategy = strategy;
        this.relativeDist = relativeDistance;
        this.parallel = parallel;
        if (mainKdTree == null) {
            throw new IllegalArgumentException("mainKdTree");
        }
        if (strategy == Strategy.POINT_TO_POINT_ABSOLUTE && relativeDistance) {
            throw new IllegalArgumentException("Point-to-point strategy cannot be used with relative distance");
        }
        this.kdTree = mainKdTree;
    }
    
    /**
     * Returns Hausdorff distance of mesh facets to the source mesh facets. 
     * Keys in the map contain mesh facets that were measued with the 
     * source facets. For each measured facet, a list of distances to the source 
     * facets is stored. The order od distances corresopnds to the order of vertices
     * in the measured facet, i.e., the i-th value is the distance of the i-th vertex
     * of the measured facet.
     * 
     * @return Hausdorff distance for all points of all measured facets
     */
    public Map<MeshFacet, List<Double>> getDistances() {
        return Collections.unmodifiableMap(distances);
    }
    
    /**
     * Returns {@code true} if the distance was computed as relative 
     * (with possibly negative values). 
     * 
     * @return {@code true} if the distance is computed as relative, {@code false} otherwise.
     */
    public boolean relativeDistance() {
        return this.relativeDist;
    }
    
    public Strategy getStrategy() {
        return strategy;
    }
    
    /**
     * Returns {@code true} if the distance computation is parallel.
     * @return {@code true} if the distance computation is parallel.
     */
    public boolean inParallel() {
        return parallel;
    }
    
    /**
     * Return the KD tree to which distances are computed.
     * 
     * @return KD tree to which distances are computed
     */
    public KdTree getMainKdTree() {
        return kdTree;
    }
    
    @Override
    protected void visitMeshFacet(MeshFacet comparedFacet) {
        if (distances.containsKey(comparedFacet)) {
            return; // multiply visited facet
        }
        
        final List<Double> distList = new ArrayList<>();
        distances.put(comparedFacet, distList);
        
        final List<MeshPoint> vertices = comparedFacet.getVertices();
        final List<Future<KdTreeVisitor>> results = new ArrayList<>(vertices.size());
        
        ExecutorService executor = Executors.newFixedThreadPool(Runtime.getRuntime().availableProcessors());
        
        for (int i = 0; i < vertices.size(); i++) { // for each vertex of comparedFacet
            KdTreeVisitor visitor = instantiateVisitor(vertices.get(i).getPosition());
            kdTree.accept(visitor); // compute the distance for i-th vertex or prepare the visitor for computation
            
            if (inParallel()) { // fork and continue
                results.add(executor.submit(visitor));
            } else { // process distance results immediately
                addDistanceToPoint(visitor, vertices.get(i), distList);
            }
        }
            
        if (inParallel()) { // process asynchronous computation of distance
            executor.shutdown();
            while (!executor.isTerminated()){
            }
            try {
                int i = 0;
                for (Future<KdTreeVisitor> res: results) {
                    final KdTreeVisitor visitor = (KdTreeVisitor) res.get(); // waits until all computations are finished
                    addDistanceToPoint(visitor, vertices.get(i), distList);
                    i++;
                }
            } catch (final InterruptedException | ExecutionException ex) {
                Logger.getLogger(HausdorffDistance.class.getName()).log(Level.SEVERE, null, ex);
            }
        }
    }
    
    protected KdTreeVisitor instantiateVisitor(Vector3d inspectedPoint) {
        switch (getStrategy()) {
            case POINT_TO_POINT_ABSOLUTE:
                return new KdTreeDistance(inspectedPoint, inParallel());
            case POINT_TO_POINT:
                return new KdTreeDistanceToVertices(inspectedPoint, inParallel());
            default:
            case POINT_TO_TRIANGLE_APPROXIMATE:
                return new KdTreeApproxDistanceToTriangles(inspectedPoint, inParallel());
        }
    }
    
    protected void addDistanceToPoint(KdTreeVisitor vis, MeshPoint point, List<Double> pointDistances) {
        if (relativeDist && vis instanceof DistanceWithNearestPoints) {
            double relDist = checkDist((DistanceWithNearestPoints)vis, point);
            pointDistances.add(relDist);
        } else {
            pointDistances.add(((Distance) vis).getDistance());
        }
    }
    
    protected double checkDist(DistanceWithNearestPoints vis, MeshPoint myPoint) {
        double dist = vis.getDistance();
        
        if (!relativeDist) {
            return dist;
        }
        
        if (vis.getNearestPoints().size() != 1) {
            return Double.POSITIVE_INFINITY;
        }
        
        MeshFacet firstFacet = vis.getNearestPoints().keySet().stream().findFirst().get();
        
        if (vis.getNearestPoints().get(firstFacet).size() != 1) {
            return Double.POSITIVE_INFINITY;
        }
        
        Vector3d firstVertex = vis.getNearestPoints().get(firstFacet).get(0);
        
        // compute sign for relative distance:
        Vector3d aux = new Vector3d(firstVertex);
        aux.sub(myPoint.getPosition());
        int sign = (int) Math.signum(aux.dot(myPoint.getNormal()));
        
        return sign * dist;
    }
}
